#!/usr/bin/python
# (c) 2009 Jiri Baum

from dbus_pprint import dbus_pprint
import getpass, sys
import dbus, gobject, telepathy
from functools import partial

from accountmgr import AccountManager
from account import Account

from telepathy.constants import (
    CONNECTION_HANDLE_TYPE_CONTACT, HANDLE_TYPE_LIST,
    CHANNEL_TEXT_MESSAGE_TYPE_NORMAL,
    TUBE_TYPE_DBUS, TUBE_STATE_LOCAL_PENDING, TUBE_STATE_OPEN,
    CONNECTION_STATUS_CONNECTED, CONNECTION_STATUS_DISCONNECTED,
    CONNECTION_PRESENCE_TYPE_OFFLINE,
    SOCKET_ACCESS_CONTROL_LOCALHOST)
from telepathy.interfaces import (
    CHANNEL, CHANNEL_INTERFACE_GROUP, CHANNEL_TYPE_CONTACT_LIST,
    CHANNEL_TYPE_DBUS_TUBE,
    CONNECTION, CONNECTION_INTERFACE_REQUESTS,
    CONNECTION_INTERFACE_CONTACTS,
    CONNECTION_INTERFACE_ALIASING,
    CONNECTION_INTERFACE_AVATARS,
    CONNECTION_INTERFACE_SIMPLE_PRESENCE,
    CONNECTION_INTERFACE_CONTACT_CAPABILITIES,
    CONNECTION_INTERFACE_CAPABILITIES,
    CHANNEL_INTERFACE_TUBE,

    CLIENT, CLIENT_HANDLER,
    ACCOUNT_MANAGER, ACCOUNT)

DBUS_PROPERTIES = 'org.freedesktop.DBus.Properties'

SERVICE = 'au.com.baum.connectr' # domain name, backwards
PATH = '/' + SERVICE.replace('.', '/')
TUBETYPE = SERVICE.replace('.', '_')

alias_key = CONNECTION_INTERFACE_ALIASING + '/alias'

# Note on the TUBETYPE: in principle, it should be equal to the SERVICE;
# however, in some places, a single word is required...

class Tubes(dbus.service.Object,
              #telepathy._generated.Client_Handler.ClientHandler,
              telepathy.server.DBusProperties):
    def __init__(self, models):
        self.models = models
        self.buddies = {}
        self.me_handles = {}
        self.views = []
        self.messages = {}
        self.connections = {}
        self.conn_display_names = {}
        self._interfaces = set([CLIENT, CLIENT_HANDLER, DBUS_PROPERTIES])
        self.service_name = CLIENT+'.'+TUBETYPE
        object_path = '/'+self.service_name.replace('.', '/')

        bus = dbus.SessionBus()

        dbus.service.Object.__init__(self, bus, object_path)
        telepathy.server.DBusProperties.__init__(self)
        self._implement_property_get(CLIENT, {
            'Interfaces': lambda: [CLIENT, CLIENT_HANDLER, DBUS_PROPERTIES],
          })
        self._implement_property_get(CLIENT_HANDLER, {
            'HandlerChannelFilter': lambda: dbus.Array([
                dbus.Dictionary({
                    CHANNEL + '.ChannelType'     : CHANNEL_TYPE_DBUS_TUBE,
                    CHANNEL + '.TargetHandleType': CONNECTION_HANDLE_TYPE_CONTACT,
                    CHANNEL_TYPE_DBUS_TUBE + '.ServiceName'    : SERVICE,
                }, signature='sv')
            ], signature='a{sv}'),
            #'Capabilities': lambda: dbus.Array([SERVICE], signature='as'),
          })
        #bus.request_name(self.service_name)
        self.name_ownership = dbus.service.BusName(self.service_name, bus=bus)

        # add a handler that watches for changes
        #bus.add_signal_receiver(partial(connections_change_cb, self),
        #    dbus_interface=CONNECTION, signal_name='StatusChanged',
        #    path_keyword='path')

        bus.add_signal_receiver(self.new_acct,
            dbus_interface=ACCOUNT_MANAGER,
            signal_name='AccountValidityChanged')
        # work through the list of connections that are on-line
        AccountManager.watch_changes(self.account_mgr_change)

    def account_mgr_change(self, new_acct_mgr):
        if not new_acct_mgr:
            self.msg('am', 'account manager',
                     'no account manager found (is Empathy running?)')
            return

        try:
            am = AccountManager()
        except:
            self.msg('am', 'account manager',
                                  "problem contacting the account manager")
            return

        self.msg('am', 'account manager') #OK, no need to show the user

        for acct_path in am[DBUS_PROPERTIES].Get(ACCOUNT_MANAGER, 'ValidAccounts'):
            self.new_acct(acct_path, True)

    def new_acct(self, acct_path, valid):
        acct = Account(acct_path)
        display_name = acct[DBUS_PROPERTIES].Get(ACCOUNT, 'DisplayName')
        if not valid:
            self.msg('account', display_name, 'Invalid.')
            return

        acct[ACCOUNT].connect_to_signal('AccountPropertyChanged',
                            partial(self.acct_changed, acct, display_name))

        self.acct_changed(acct, display_name, {})

    def acct_changed(self, acct, display_name, properties):
        conn_path = acct[DBUS_PROPERTIES].Get(ACCOUNT, 'Connection')
        if conn_path == '/':
            self.msg('account', display_name, 'Account offline.')
            return

        if conn_path not in self.connections:
            self.connections[conn_path] = None
            self.msg('account', display_name, 'Contacting connection mgr.')
            conn_name = conn_path.replace('/', '.')[1:]
            self.conn_display_names[conn_path] = display_name
            conn = telepathy.client.Connection(conn_name, conn_path)
            self.connections[conn_path] = conn
            conn.call_when_ready(self.ready_cb)

    def ready_cb(self, conn):
        """ Called when the connection is ready... """

        if conn.object_path not in self.buddies:
            self.buddies[conn.object_path] = {}

        self.msg('account', self.conn_display_names[conn.object_path],
                                                 'Retrieving my own info.')

        # link up the AvatarRetrieved signal now (rather than in
        # get_roster) so that it's ready to catch the self-avatar
        conn[CONNECTION_INTERFACE_AVATARS].connect_to_signal(
            'AvatarRetrieved', partial(self.avatar_retrieved, conn))

        # get the self handle (so I know my own avatar):
        conn[CONNECTION].connect_to_signal(
            'SelfHandleChanged',
            partial(self.self_handle_changed, conn))
        me = conn[DBUS_PROPERTIES].Get(CONNECTION, 'SelfHandle')
        self.me_handles[conn.object_path] = me
        self.got_new_buddies(conn, [me])
        me_information.set_buddy(self.buddies[conn.object_path][me])

        self.msg('account', self.conn_display_names[conn.object_path],
                                                  'Retrieving buddy list.')

        # get list of buddies:
        conn[CONNECTION_INTERFACE_CONTACT_CAPABILITIES].connect_to_signal(
            'ContactCapabilitiesChanged',
            partial(self.contact_capabilities_changed, conn))
        for group in ('subscribe', 'publish', 'stored', 'known'):
            conn[CONNECTION_INTERFACE_REQUESTS].EnsureChannel({
                CHANNEL + '.ChannelType'     : CHANNEL_TYPE_CONTACT_LIST,
                CHANNEL + '.TargetHandleType': HANDLE_TYPE_LIST,
                CHANNEL + '.TargetID'        : group,
                },
                reply_handler = partial(self.member_channel_cb, conn),
                error_handler = partial(self.roster_error_cb, group))

        # buddies will be contacted as their details come in...

    def self_handle_changed(self, conn, new_handle):
        print "TODO: handle the self handle changing (signal ignored...)"


    def avatar_retrieved(self, conn, buddy_h, token, avatar, type):
        buddy = self.buddies[conn.object_path][buddy_h]
        buddy['avatar'] = (type, avatar)

        for cb in buddy['avatar_queue']:
            cb(type, avatar)
        buddy['avatar_queue'] = []

    def roster_error_cb(self, group, *args):
        if group=='known':
            # the "known contacts" list is obsolete, don't mention it
            pass
            #print 'Info: could not retrieve the obsolete "known contacts" list'
            #print 'The error was:', args
        else:
            print 'Warning: could not retrieve the "%s contacts" list'%group
            print 'The error was:', args

    def member_channel_cb(self, conn, yours, path, properties):
        channel = telepathy.client.Channel(conn.service_name, path)
        channel[CHANNEL_INTERFACE_GROUP].connect_to_signal(
                  'MembersChanged', partial(self.members_changed_cb, conn))

        handles = channel[DBUS_PROPERTIES].Get(CHANNEL_INTERFACE_GROUP,
                                                                 'Members')

        handles = set(handles)
        if handles:
            self.got_new_buddies(conn, handles)
            self.msg('account', self.conn_display_names[conn.object_path],
                                                  'OK')


    def got_new_buddies(self, conn, handles):
        attribs = conn[CONNECTION_INTERFACE_CONTACTS].GetContactAttributes(
                handles, [
                    CONNECTION,
                    CONNECTION_INTERFACE_ALIASING,
                    CONNECTION_INTERFACE_SIMPLE_PRESENCE,
                    CONNECTION_INTERFACE_CONTACT_CAPABILITIES,
                ],
                False)
        for h, details in attribs.items():
            if h in self.buddies[conn.object_path]:
                self.buddies[conn.object_path][h].update(details)
            else:
                self.buddies[conn.object_path][h] = details
        self.select_buddy(conn, handles)
        #else:
        #    alias_key = CONNECTION_INTERFACE_ALIASING + '/alias'
        #    buddy = self.buddies[conn.object_path][handles[0]]
        #    buddy_name = buddy[alias_key]
        #    print "I am %s"%buddy_name

    def contact_capabilities_changed(self, conn, changed_caps):
        # notes:
        #   1. We're only handling here the situation where contacts go
        #   online; going offline is also signalled here...
        #   2. We should check whether changed_caps contains the relevant
        #   capabilities, to save on re-retrieving irrelevant data.
        self.got_new_buddies(conn, changed_caps.keys())

    def members_changed_cb(self, conn, message, added, removed, local_pending,
            remote_pending, actor, reason):

        # Note: new buddies are signalled here, but not buddies that were
        # offline and have now come online (or vice versa).

        if added:
            self.got_new_buddies(conn, added)
        if removed:
            for handle in removed:
                member = self.conn[CONNECTION].InspectHandles(
                    CONNECTION_HANDLE_TYPE_CONTACT, [handle])[0]
                print 'buddy went away: %s' % member

    def select_buddy(self, conn, new_buddies, me=False):
        presence_key = CONNECTION_INTERFACE_SIMPLE_PRESENCE + '/presence'
        found_buddies = []
        for buddy_h in new_buddies:
            buddy = self.buddies[conn.object_path][buddy_h]
            buddy_name = buddy[alias_key]
            self.msg('buddy', buddy_name, "???")
            me = buddy_h == self.me_handles[conn.object_path]
            caps = (not me) and self.check_buddy_caps(conn, buddy_h)
            if me:
                self.msg('buddy', buddy_name, "That's me!")
            if caps:
                self.msg('buddy', buddy_name, "Contacting!")
                self.contact(conn, buddy_h)
            if caps or me:
                if 'avatar_queue' not in buddy and 'avatar' not in buddy:
                    buddy['avatar_queue'] = []
                    found_buddies.append(buddy_h)
            elif caps is None:
                pres=self.buddies[conn.object_path][buddy_h][presence_key]
                if pres[0]==CONNECTION_PRESENCE_TYPE_OFFLINE:
                    self.msg('buddy',
                        self.buddies[conn.object_path][buddy_h][alias_key],
                        "Offline.")
                else:
                    if pres[2]:
                        status = '%s - %s'%(pres[1], pres[2])
                    else:
                        status = pres[1]

                    self.msg('buddy',
                        self.buddies[conn.object_path][buddy_h][alias_key],
                        "No capabilities obtained, status %s."%status)
            else:
                self.msg('buddy', buddy_name, "Can't handle %s."%(SERVICE))

        if found_buddies:
            conn[CONNECTION_INTERFACE_AVATARS].RequestAvatars(found_buddies)

    def check_buddy_caps(self, conn, buddy_h):
        caps = CONNECTION_INTERFACE_CONTACT_CAPABILITIES + '/caps'
        if caps not in self.buddies[conn.object_path][buddy_h]:
            return None
        for props, additional in self.buddies[conn.object_path][buddy_h][caps]:
            if props.get(CHANNEL_TYPE_DBUS_TUBE + '.ServiceName', '')==SERVICE:
                return True
        return False

    def contact(self, conn, buddy_h):
        conn[CONNECTION_INTERFACE_REQUESTS].EnsureChannel({
            CHANNEL + '.ChannelType'     : CHANNEL_TYPE_DBUS_TUBE,
            CHANNEL + '.TargetHandleType': CONNECTION_HANDLE_TYPE_CONTACT,
            CHANNEL + '.TargetHandle'    : buddy_h,
            CHANNEL_TYPE_DBUS_TUBE + '.ServiceName'    : SERVICE,
            },
            reply_handler = partial(self.tube_channel_cb, conn, buddy_h),
            error_handler = self.error_cb)

    def tube_channel_cb(self, conn, buddy_h, yours, path, properties):
        channel = telepathy.client.Channel(conn.service_name, path)
        addr = channel[CHANNEL_TYPE_DBUS_TUBE].Offer({}, SOCKET_ACCESS_CONTROL_LOCALHOST)
        channel[CHANNEL_INTERFACE_TUBE].connect_to_signal('TubeChannelStateChanged',
                    partial(self.tube_state_cb, conn, addr, buddy_h))

    def tube_state_cb(self, conn, addr, buddy_h, state):
        if state == TUBE_STATE_OPEN:
            tube = dbus.connection.Connection(addr)
            #tube.add_signal_receiver(self.signal_cb)
            me = ConnectrObj(tube, conn)
            other = tube.get_object(object_path=PATH)
            buddy_info = BuddyInfo(self.buddies[conn.object_path][buddy_h])
            self.msg('buddy', buddy_info.name(), 'Connected.')

            for model in self.models:
                model.subscribe_tube(tube, me, other, buddy_info)

    def use_connection(self, me, other):
        me.Hello('hello from %s'%getpass.getuser())
        other.Method("xyzzy %s"%getpass.getuser(),
                  reply_handler=self.reply_cb, error_handler=self.error_cb)

    def signal_cb(self, *args, **kwargs):
        print 'Signal:', args, kwargs

    def reply_cb(self, *args):
        print 'Reply:', args

    def error_cb(self, *args):
        print 'Error:', args

    @dbus.service.method(dbus_interface=CLIENT_HANDLER,
                           in_signature='ooa(oa{sv})aota{sv}', out_signature='')
    def HandleChannels(self, acct, conn_path, channels, reqs_satisfied,
                                           user_action_time, handler_info):
        #print 'incoming channel'
        conn_name = conn_path.replace('/', '.')[1:]
        conn = telepathy.client.Connection(conn_name, conn_path)
        for path, props in channels:
            assert props[CHANNEL + '.ChannelType'] == CHANNEL_TYPE_DBUS_TUBE
            assert props[CHANNEL_TYPE_DBUS_TUBE + '.ServiceName'] == SERVICE
            assert props[CHANNEL + '.TargetHandleType'] == CONNECTION_HANDLE_TYPE_CONTACT
            assert not props[CHANNEL + '.Requested']
            channel = telepathy.client.Channel(conn_name, path)
            addr = channel[CHANNEL_TYPE_DBUS_TUBE].Accept(SOCKET_ACCESS_CONTROL_LOCALHOST)
            channel[CHANNEL_INTERFACE_TUBE].connect_to_signal(
              'TubeChannelStateChanged', partial(self.tube_state_cb, conn,
                                  addr, props[CHANNEL + '.TargetHandle'] ))
            if len(handler_info):
                print 'handler info:',
                dbus_pprint(handler_info)

    def subscribe_view(self, view):
        self.views.append(view)

    def unsubscribe_view(self, view):
        self.views.remove(view)

    def msg(self, typ, name, msg=None):
        #print typ, name, msg
        key = (typ, str(name))
        if msg is None:
            if key in self.messages:
                del self.messages[key]
        else:
            self.messages[key]=msg

        for v in self.views:
            v.update()

class ConnectrObj(dbus.service.Object):
    def __init__(self, tube, conn):
        super(ConnectrObj, self).__init__(tube, PATH)
        self.tube = tube

    @dbus.service.signal(dbus_interface=SERVICE, signature='sd')
    def StatusUpdate(self, txt, when):
        pass

    @dbus.service.method(dbus_interface=SERVICE, in_signature='',
                                                         out_signature='i')
    def ProtoVersion(self):
        return 1

class BuddyInfo():
    def __init__(self, buddy):
        self.buddy = buddy

    def name(self):
        return self.buddy[alias_key]

    def avatar_rq(self, cb):
        if 'avatar' in self.buddy:
            cb(*self.buddy['avatar'])
        elif 'avatar_queue' in self.buddy:
            self.buddy['avatar_queue'].append(cb)
        else:
            raise KeyError, "that buddy doesn't have avatars coming"

class MeInfo(BuddyInfo):
    def __init__(self):
        self.buddy = None
        self.queue = []

    def avatar_rq(self, cb):
        if self.buddy is None:
            self.queue.append(cb)
        else:
            BuddyInfo.avatar_rq(self, cb)

    def set_buddy(self, buddy):
        self.buddy = buddy
        for cb in self.queue:
            self.avatar_rq(cb)
        self.queue = []

me_information = MeInfo()

if __name__ == '__main__':
    import dbus.glib # sets up a main loop (!!!); somewhat deprecated

    ct = Tubes()

    try:
        gobject.MainLoop().run()
    except KeyboardInterrupt:
        print 'interrupted'

